"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getTokenExpirationTimeout = exports.setTokenExpirationListener = exports.getRequestedPage = exports.getActivePageSize = exports.powerbi = void 0;
const powerbi_client_1 = require("powerbi-client");
const version_1 = require("./version");
// Jupyter SDK type to be passed with service instance creation
const JUPYTER_SDK_TYPE = 'powerbi-jupyter';
// Initialize powerbi service
exports.powerbi = new powerbi_client_1.service.Service(powerbi_client_1.factories.hpmFactory, powerbi_client_1.factories.wpmpFactory, powerbi_client_1.factories.routerFactory, { type: JUPYTER_SDK_TYPE, sdkWrapperVersion: version_1.MODULE_VERSION });
// Set threshold to refresh token in minutes
const TOKEN_REFRESH_THRESHOLD = 10;
/**
 * Returns width and height of the active page as an object
 * @param report PowerBI Report
 */
function getActivePageSize(report) {
    return __awaiter(this, void 0, void 0, function* () {
        const pages = yield report.getPages();
        // Get the active page
        const activePage = pages.find((page) => {
            return page.isActive;
        });
        if (!activePage) {
            throw 'No active report page';
        }
        return activePage.defaultSize;
    });
}
exports.getActivePageSize = getActivePageSize;
function getRequestedPage(report, pageName) {
    return __awaiter(this, void 0, void 0, function* () {
        const pages = yield report.getPages();
        const requestedPage = pages.filter((page) => {
            return page.name === pageName;
        })[0];
        if (!requestedPage) {
            throw 'Page not found';
        }
        return requestedPage;
    });
}
exports.getRequestedPage = getRequestedPage;
/**
 * Set token expiration listener
 * @param accessToken
 * @param widgetView ReportView | QuickVisualizeView
 */
function setTokenExpirationListener(accessToken, widgetView) {
    const timeout = getTokenExpirationTimeout(accessToken);
    // If timeout reaches safetyInterval, invoke the setTokenExpiredFlag method
    if (timeout <= 0) {
        widgetView.setTokenExpiredFlag();
    }
    else {
        // Set timeout so minutesToRefresh minutes before token expires, setTokenExpiredFlag will be invoked
        console.log('Access Token will expire in ' + timeout + ' milliseconds.');
        setTimeout(() => {
            widgetView.setTokenExpiredFlag();
        }, timeout);
    }
}
exports.setTokenExpirationListener = setTokenExpirationListener;
/**
 * Parse the given token and get the expiration timeout
 * @param token
 */
function getTokenExpirationTimeout(token) {
    const tokenExpiration = tryGetTokenExpiration(token);
    if (!tokenExpiration) {
        return 0;
    }
    // Get current time
    const currentTime = Date.now();
    const expiration = tokenExpiration * 1000;
    const safetyInterval = TOKEN_REFRESH_THRESHOLD * 60 * 1000;
    // Time until token refresh in milliseconds
    return expiration - currentTime - safetyInterval;
}
exports.getTokenExpirationTimeout = getTokenExpirationTimeout;
// window.atob() won't decode the access token with unicodes properly. So, we are using from bytestream, to percent-encoding, to original string.
// reference: https://developer.mozilla.org/en-US/docs/Web/API/WindowBase64/Base64_encoding_and_decoding#The_Unicode_Problem
// reference: https://stackoverflow.com/questions/38552003/how-to-decode-jwt-token-in-javascript-without-using-a-library/38552302#38552302
function parseJSONWebToken(token) {
    let payload;
    const encodedPayload = token.split('.')[1];
    // Fix for parsing the access token with unicode characters Bug: 261401
    payload = encodedPayload.replace(/-/g, '+').replace(/_/g, '/');
    var jsonPayload = decodeURIComponent(atob(payload).split('').map(function (c) {
        return '%' + ('00' + c.charCodeAt(0).toString(16)).slice(-2);
    }).join(''));
    return JSON.parse(jsonPayload);
}
/**
 * Parse the given token and get the expiration timestamp
 * @param token
 */
function tryGetTokenExpiration(token) {
    if (!token) {
        return undefined;
    }
    try {
        const tokenDetails = parseJSONWebToken(token);
        return tokenDetails === null || tokenDetails === void 0 ? void 0 : tokenDetails.exp;
    }
    catch (ex) {
        return undefined;
    }
}
//# sourceMappingURL=utils.js.map